/** @file floppy.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.11.05 -

	@brief [ floppy drive ]
*/

#ifndef _FLOPPY_H_
#define _FLOPPY_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "parsewav.h"

// #define _DEBUG_FLOPPY

#define SIG_FLOPPY_ACCESS			0
#define SIG_FLOPPY_READ_ID			1
//#define SIG_FLOPPY_READ_ID_TRACK_NUM		2
//#define SIG_FLOPPY_READ_ID_HEAD_NUM		3
#define SIG_FLOPPY_READ				6
#define SIG_FLOPPY_READ_TRACK		7
#define SIG_FLOPPY_WRITE			8
#define SIG_FLOPPY_WRITE_TRACK		9
#define SIG_FLOPPY_WRITEDELETE		10
#define SIG_FLOPPY_WRITEPROTECT		11
#define SIG_FLOPPY_STEP				12	// step and dirc
#define SIG_FLOPPY_HEADLOAD			13
#define SIG_FLOPPY_READY			14
#define SIG_FLOPPY_TRACK0			15
#define SIG_FLOPPY_INDEX			16
#define SIG_FLOPPY_DELETED			17

#define SIG_FLOPPY_CURRENTTRACK		18
#define SIG_FLOPPY_SECTOR_NUM		19
#define SIG_FLOPPY_SECTOR_SIZE		20
#define SIG_FLOPPY_TRACK_SIZE		21

// for 5inch mini floppy
#define SIG_FLOPPY_IRQ				26
#define SIG_FLOPPY_DRQ				27

#define FLOPPY_MAX_EVENT_V1			5
#define FLOPPY_MAX_EVENT			20

//class FDC;
class DISK;

/**
	@brief floppy drive
*/
class FLOPPY : public DEVICE
{
private:
	DEVICE *d_fdc3, *d_fdc5, *d_fdc5ex, *d_board;

	// config
	bool ignore_crc;

	// event
	int register_id[FLOPPY_MAX_EVENT];

	//
	uint8 drv_num[2];	// drive number
	uint8 drvsel[2];	// drive select + motor

	uint8 sidereg;	// head select

	uint8 index_hole;	// index hole
	uint8 head_load;	// head loaded

	uint8 density;	// density (double = 1)
	uint8 motor_on_expand[2];	// motor on (expand)

	int sectorcnt;

	uint8 fdd5outreg[2];
	bool irqflg;
	bool irqflgprev;
	bool drqflg;
	bool drqflgprev;

	// output signals
	outputs_t outputs_irq;
	outputs_t outputs_drq;

	// drive info
	typedef struct {
		int side;
		int track;
		int index;
		bool access;
		uint8 ready;	// ready warmup:01 on:11
		uint8 motor_warmup;	// motor warming up:1
		uint8 head_loading;
	} fdc_t;
	fdc_t fdc[MAX_DRIVE];

	// diskette info
	DISK* disk[MAX_DRIVE];

	// index hole
	int delay_index_hole;
	// ready on
	int delay_ready_on;

	bool ff00_mask;

	int sample_rate;
	int max_vol;
	int wav_volume;
	int wav_enable;
	bool wav_loaded_at_first;

	// sound data 1st dimension [0]:3inch [1]:5inch [2]:5inch2HD
#define FLOPPY_WAV_FDDTYPES 3
	// 2nd [0]:seek sound [1]:motor sound [2]:head on [3]:head off
#define FLOPPY_WAV_SNDTYPES 4
#define FLOPPY_WAV_SEEK		0
#define FLOPPY_WAV_MOTOR	1
#define FLOPPY_WAV_HEADON	2
#define FLOPPY_WAV_HEADOOFF	3
	_TCHAR wav_file[FLOPPY_WAV_FDDTYPES][FLOPPY_WAV_SNDTYPES][20];
	uint8 *wav_data[FLOPPY_WAV_FDDTYPES][FLOPPY_WAV_SNDTYPES];
	int    wav_size[FLOPPY_WAV_FDDTYPES][FLOPPY_WAV_SNDTYPES];
	// [0]:seek sound [1]:motor sound [2]:head on [3]:head off
	int    wav_play_pos[FLOPPY_WAV_SNDTYPES];
	int    wav_play[FLOPPY_WAV_SNDTYPES];

	//for resume
#pragma pack(1)
	struct vm_state_st_v1 {
		int register_id[6];
		uint8 drv_num;
		uint8 drvsel;
		uint8 sidereg;
		uint8 fdd5outreg;
		uint8 irqflg;
		uint8 drqflg;

		char reserved[2];
	};
	struct vm_state_st_v2 {
		int register_id[22];
		uint8 drv_num;
		uint8 drvsel;
		uint8 sidereg;
		uint8 fdd5outreg;
		uint8 irqflg;
		uint8 drqflg;

		// version 3
		uint8 drv_num2;
		uint8 drvsel2;

		uint8 fdd5outreg2;

		// version 0x43
		uint8 ff00_mask;

		char reserved[14];
	};
#pragma pack()
	// cannot write when load state from file
	bool ignore_write;

	void warm_reset();

	inline void cancel_my_event(int);
	inline void register_my_event(int, int);
	void cancel_my_events();

	void set_drive_select(int type, uint8 data);

	// irq/dma
	void set_irq(bool val);
	void set_drq(bool val);

	void load_wav();

	void motor(int drv, bool val);

	void set_drive_speed();

public:
	FLOPPY(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("FLOPPY");
		init_output_signals(&outputs_irq);
		init_output_signals(&outputs_drq);
		d_fdc3 = NULL;
		d_fdc5 = NULL;
	}
	~FLOPPY() {}

	// common functions
	void initialize();
	void reset();
	void release();
	void write_io8(uint32 addr, uint32 data);
	uint32 read_io8(uint32 addr);
	void write_signal(int id, uint32 data, uint32 mask);
	uint32 read_signal(int id);

	// unique functions
	void set_context_fdc(DEVICE* device3, DEVICE* device5, DEVICE* device5e) {
		d_fdc3 = device3;
		d_fdc5 = device5;
		d_fdc5ex = device5e;
	}
	void set_context_irq(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_irq, device, id, mask);
	}
	void set_context_drq(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_drq, device, id, mask);
	}
	void set_context_board(DEVICE* device) {
		d_board = device;
	}

	// event callback
	void event_frame();
	void event_callback(int event_id, int err);

	bool search_track(int channel);
	bool verify_track(int channel);
	int  search_sector(int channel);
	int  search_sector(int channel, int sect);
	int  search_sector(int channel, int sect, bool compare_side, int side);
	bool make_track(int channel);
	bool parse_track(int channel);

	// user interface
	bool open_disk(int drv, const _TCHAR *path, int offset, uint32 flags);
	void close_disk(int drv);
	int  change_disk(int drv);
	void set_disk_side(int drv, int side);
	int  get_disk_side(int drv);
	bool disk_inserted(int drv);
	void set_drive_type(int drv, uint8 type);
	uint8 get_drive_type(int drv);
	uint8 fdc_status();
	void toggle_disk_write_protect(int drv);
	bool disk_write_protected(int drv);

	uint16 get_drive_select();

	void mix(int32* buffer, int cnt);
	void set_volume(int vol);
	void init(int rate, int volume);

	void save_state(FILEIO *fp);
	void load_state(FILEIO *fp);

#ifdef USE_DEBUGGER
	uint32 debug_read_io8(uint32 addr);
	bool debug_write_reg(uint32 reg_num, uint32 data);
	void debug_regs_info(_TCHAR *buffer, size_t buffer_len);
#endif
};

#endif

